﻿using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;
using Cinemachine;

namespace Bonfire
{
    public class PlayerMove : MonoBehaviour
    {
        [SerializeField] private Text debugText = null;
        [SerializeField] private Transform lookTarget = null;
        [SerializeField] Transform lockOnTarget = null;

        // 重力値
        const float GravityPower = 9.8f;

        // 移動速度
        internal float walkSpeed = 1f;
        internal float runSpeedUpRatio = 2.0f;
        internal float sprintSpeedUpRatio = 2.0f;

        // 回転速度
        const float rotationSpeed = 1080.0f;

        // ジャンプの強さ
        const float jumpPower = 10f;
        // あまり速いと地面を突き抜けそうな気がするので、最高落下速度を決めておく。
        const float maxFallSpeed = -30f;

        // 現在の速度 (実際に使用するのは y 方向のみ)
        Vector3 currentVelocity = Vector3.zero;

        // 着地しているかどうか
        bool isGrounded = false;

        // 着地判定に飛ばす Ray の距離
        [SerializeField] float isGroundedTestRayDistance = 0.15f;

        // ジャンプ中かどうかの判定
        bool isJumpStarted = false;

        Camera mainCamera = null;
        Animator animator = null;
        CharacterController characterController = null;
        PlayerCtrl playerCtrl = null;
        PlayerStatus status = null;
        CinemachineFreeLook freeLookCamera = null;
        CinemachineVirtualCamera virtualCamera = null;

        void Start()
        {
            mainCamera = FindObjectOfType<Camera>();
            animator = GetComponent<Animator>();
            characterController = GetComponent<CharacterController>();
            playerCtrl = GetComponent<PlayerCtrl>();
            status = GetComponent<PlayerStatus>();

            freeLookCamera = FindObjectOfType<CinemachineFreeLook>();
            foreach (var vcam in FindObjectsOfType<CinemachineVirtualCamera>())
            {
                if (vcam.name == "CM vcam1")
                {
                    virtualCamera = vcam;
                }
            }
        }

        void Update()
        {
            // 移動速度
            Vector3 velocity = Vector3.zero;

            float moveSpeed = 0f;
            float speedByStick = 0f;
            float runSpeedUp = 1.0f;
            float sprintSpeedUp = 1.0f;

            // 独自の着地判定
            Ray ray = new Ray(transform.position + 0.18f * transform.up, -transform.up);
            isGrounded = Physics.SphereCast(ray, 0.13f, isGroundedTestRayDistance);
            status.isGrounded = isGrounded;

            if (status.isLockOn)
            {
                freeLookCamera.transform.position = mainCamera.transform.position;

                if (lockOnTarget != null)
                {
                    if (playerCtrl.canRotate)
                    {
                        Quaternion characterTargetRotation = Quaternion.LookRotation(lockOnTarget.position - transform.position);
                        // 強引だがこれで Y 軸の回転だけにする
                        characterTargetRotation.x = 0;
                        characterTargetRotation.z = 0;
                        transform.rotation = Quaternion.RotateTowards(transform.rotation, characterTargetRotation, rotationSpeed * Time.deltaTime);
                    }
                }
                else
                {
                    status.isLockOn = false;
                }
            }
            else
            {
                if (playerCtrl.IsInputStickL())
                {
                    // プレイヤーの向き
                    Vector3 playerDirection = playerCtrl.GetPlayerDirection();

                    // カメラの向きに、スティックの向きを掛け合わせ、キャラクターをスティックの方向に向かせる
                    Vector3 cameraDirection = lookTarget.position - mainCamera.transform.position;
                    cameraDirection.y = 0;
                    cameraDirection = Vector3.Normalize(cameraDirection);
                    if (playerCtrl.canRotate)
                    {
                        Quaternion characterTargetRotation = Quaternion.LookRotation(cameraDirection) * Quaternion.LookRotation(playerDirection);
                        transform.rotation = Quaternion.RotateTowards(transform.rotation, characterTargetRotation, rotationSpeed * Time.deltaTime);
                    }
                }
            }

            // x ボタンでの加速
            if (playerCtrl.isRunning)
            {
                runSpeedUp = runSpeedUpRatio;
            }
            else
            {
                playerCtrl.isSprint = false;
            }
            // L3 ボタンでのsprint加速
            if (playerCtrl.isSprint)
            {
                sprintSpeedUp = sprintSpeedUpRatio;
            }

            // スティックの傾きによる速さの変化
            speedByStick = playerCtrl.GetMagunitudeStickL();

            moveSpeed = speedByStick * walkSpeed * runSpeedUp * sprintSpeedUp;
            if (!playerCtrl.canMove)
            {
                // 動けない場合に固定値を使用する
                moveSpeed = playerCtrl.fixedMoveSpeed;
            }

            if (status.isLockOn)
            {
                Vector3 stickDirection = playerCtrl.GetPlayerDirection();
                float h = stickDirection.x;
                float v = stickDirection.z;
                velocity = transform.right * h + transform.forward * v;
                animator.SetFloat("MoveRight", h * moveSpeed);
                animator.SetFloat("MoveForward", v * moveSpeed);
                velocity = velocity * moveSpeed;
            }
            else
            {
                velocity = transform.forward.normalized * moveSpeed;
            }

            // 重力
            currentVelocity.y -= GravityPower * Time.deltaTime;
            // 速くなりすぎないようにする（速すぎると地面を突き抜けやすくなりそうな気がするため）
            currentVelocity.y = Mathf.Max(currentVelocity.y, maxFallSpeed);

            // 主にジャンプの着地判定
            if (isGrounded && currentVelocity.y < -0.1f && isJumpStarted)
            {
                status.resetStatus();
                isJumpStarted = false;
            }
            if (isGrounded && !isJumpStarted)
            {
                // 接地していたら思いっきり地面に押し付ける
                currentVelocity.y = -1;
            }

            // Y 方向の速度を加える
            velocity += currentVelocity;

            // アニメーションの速度設定
            animator.SetFloat("Speed", moveSpeed);
            animator.SetFloat("SpeedVertical", velocity.y);

            // CharacterControllerを使って動かす.
            characterController.Move(velocity * Time.deltaTime);

            // Debug 
            if (debugText != null)
            {
                debugText.text =
                   String.Format("isGrounded: {0}\ncc.isGrounded: {1}\nSpeed: {2}\nvelocity.y: {3: 0; -0}\ncurrentVelocity: {4}\npos: {5}\nisJumpStarted: {6}",
                                isGrounded,
                                characterController.isGrounded,
                                moveSpeed,
                                velocity.y,
                                currentVelocity,
                                transform.position.y,
                                isJumpStarted
                               );
            }
        }

        public void AddJumpPower()
        {
            isJumpStarted = true;
            currentVelocity.y = jumpPower;
        }

        public void ChangeLockOnCamera()
        {
            if (status.isLockOn)
            {
                freeLookCamera.m_BindingMode = CinemachineTransposer.BindingMode.LockToTarget;
                virtualCamera.Priority = 12;
            }
            else
            {
                freeLookCamera.m_BindingMode = CinemachineTransposer.BindingMode.SimpleFollowWithWorldUp;
                virtualCamera.Priority = 8;
            }
        }

        // Call by Unity
        private void OnDrawGizmosSelected()
        {
            // 着地判定の範囲を可視化する
            Gizmos.color = Color.blue;
            Gizmos.DrawWireSphere(transform.position + (0.18f - isGroundedTestRayDistance) * transform.up, 0.13f);
        }
    }

}

